# SPDX-FileCopyrightText: 2023 Max Sikström
# SPDX-License-Identifier: MIT

# Copyright © 2023 Max Sikström
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import gdb
import argparse
from .common import *


class ArmToolsSysTick (gdb.Command):
    """Dump of ARM SysTick"""

    # https://developer.arm.com/documentation/dui0552/a/cortex-m3-peripherals/system-timer--systick
    regs = [
        RegisterDef("SYST_CSR", "SysTick Control and Status Register", 0xE000E010, 4, [
            FieldBitfield("ENABLE", 0, 1),
            FieldBitfield("TICKINT", 1, 1),
            FieldBitfield("CLKSOURCE", 2, 1),
            FieldBitfield("COUNTFLAG", 16, 1)
        ]),
        RegisterDef("SYST_RVR", "SysTick Reload Value Register", 0xE000E014, 4, [
            FieldBitfield("RELOAD", 0, 24)
        ]),
        RegisterDef("SYST_CVR", "SysTick Current Value Register", 0xE000E018, 4, [
            FieldBitfield("CURRENT", 0, 24)
        ]),
        RegisterDef("SYST_CALIB", "SysTick Calibration Value Register", 0xE000E01C, 4, [
            FieldBitfield("TENMS", 0, 25),
            FieldBitfield("SKEW", 30, 1),
            FieldBitfield("NOREF", 31, 1)
        ]),
    ]

    def __init__(self):
        super().__init__('arm systick', gdb.COMMAND_USER)
        self.parser = argparse.ArgumentParser(
            description="Inspect SysTick"
        )
        self.parser.add_argument(
            '-d', '--descr',
            dest='descr',
            action='store_const',
            const=True,
            default=False,
            help="Include description"
        )

    def invoke(self, argument, from_tty):
        argv = gdb.string_to_argv(argument)
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            # We're running argparse in gdb, don't exit just return
            return

        inf = gdb.selected_inferior()
        for reg in self.regs:
            reg.dump(inf, args.descr)



# SPDX-FileCopyrightText: 2023 Max Sikström
# SPDX-License-Identifier: MIT

# Copyright © 2023 Max Sikström
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import gdb
import argparse
from .common import *


class ArmToolsSCB (gdb.Command):
    """Dump of ARM SCB"""

    # https://developer.arm.com/documentation/dui0552/a/cortex-m3-peripherals/system-control-block
    regs = [
        RegisterDef("ACTLR", "Auxiliary Control Register", 0xE000E008, 4, [
            FieldBitfield("DISMCYCINT", 0, 1),
            FieldBitfield("DISDEFWBUF", 1, 1),
            FieldBitfield("DISFOLD", 2, 1)
        ]),
        RegisterDef("CPUID", "CPUID Base Register", 0xE000ED00, 4, [
            FieldBitfield("Revision", 0, 4),
            FieldBitfield("PartNo", 4, 12),
            FieldBitfield("Constant", 16, 4),
            FieldBitfield("Variant", 20, 4),
            FieldBitfield("Implementer", 24, 8)
        ]),
        RegisterDef("ICSR", "Interrupt Control and State Register", 0xE000ED04, 4, [
            FieldBitfield("VECTACTIVE", 0, 8),
            FieldBitfield("RETTOBASE", 11, 1),
            FieldBitfield("VECTPENDING", 12, 6),
            FieldBitfield("ISRPENDING", 22, 1),
            FieldBitfield("Reserved for Debug use", 23, 1),
            FieldBitfield("PENDSTCLR", 25, 1),
            FieldBitfield("PENDSTSET", 26, 1),
            FieldBitfield("PENDSVSET", 28, 1),
            FieldBitfield("NMIPENDSET", 31, 1)
        ]),
        RegisterDef("VTOR", "Vector Table Offset Register", 0xE000ED08, 4, [
            FieldBitfield("TBLOFF", 7, 25)
        ]),
        RegisterDef("AIRCR", "Application Interrupt and Reset Control Register", 0xE000ED0C, 4, [
            FieldBitfield("VECTRESET", 0, 1),
            FieldBitfield("VECTCLRACTIVE", 1, 1),
            FieldBitfield("SYSRESETREQ", 2, 1),
            FieldBitfield("PRIGROUP", 8, 3),
            FieldBitfield("ENDIANNESS", 15, 1),
            FieldBitfield("VECTKEY", 16, 16)
        ]),
        RegisterDef("SCR", "System Control Register", 0xE000ED10, 4, [
            FieldBitfield("SLEEPONEXIT", 1, 1),
            FieldBitfield("SLEEPDEEP", 2, 1),
            FieldBitfield("SEVONPEND", 4, 1)
        ]),
        RegisterDef("CCR", "Configuration and Control Register", 0xE000ED14, 4, [
            FieldBitfield("NONBASETHRDENA", 0, 1),
            FieldBitfield("USERSETMPEND", 1, 1),
            FieldBitfield("UNALIGN_TRP", 3, 1),
            FieldBitfield("DIV_0_TRP", 4, 1),
            FieldBitfield("BFHFNMIGN", 8, 1),
            FieldBitfield("STKALIGN", 9, 1)
        ]),
        RegisterDef("SHPR1", "System Handler Priority Register 1", 0xE000ED18, 4, [
            FieldBitfield("PRI_4 - MemManage", 0, 8),
            FieldBitfield("PRI_5 - BusFault", 8, 8),
            FieldBitfield("PRI_6 - UsageFault", 16, 8),
            FieldBitfield("PRI_7", 24, 8)
        ]),
        RegisterDef("SHPR2", "System Handler Priority Register 2", 0xE000ED1C, 4, [
            FieldBitfield("PRI_8", 0, 8),
            FieldBitfield("PRI_9", 8, 8),
            FieldBitfield("PRI_10", 16, 8),
            FieldBitfield("PRI_11 - SVCall", 24, 8)
        ]),
        RegisterDef("SHPR3", "System Handler Priority Register 3", 0xE000ED20, 4, [
            FieldBitfield("PRI_12", 0, 8),
            FieldBitfield("PRI_13", 8, 8),
            FieldBitfield("PRI_14 - PendSV", 16, 8),
            FieldBitfield("PRI_15 - SysTick", 24, 8)
        ]),
        RegisterDef("SHCRS", "System Handler Control and State Register", 0xE000ED24, 4, [
            FieldBitfield("MEMFAULTACT", 0, 1),
            FieldBitfield("BUSFAULTACT", 1, 1),
            FieldBitfield("USGFAULTACT", 3, 1),
            FieldBitfield("SVCALLACT", 7, 1),
            FieldBitfield("MONITORACT", 8, 1),
            FieldBitfield("PENDSVACT", 10, 1),
            FieldBitfield("SYSTICKACT", 11, 1),
            FieldBitfield("USGFAULTPENDED", 12, 1),
            FieldBitfield("MEMFAULTPENDED", 13, 1),
            FieldBitfield("BUSFAULTPENDED", 14, 1),
            FieldBitfield("SVCALLPENDED", 15, 1),
            FieldBitfield("MEMFAULTENA", 16, 1),
            FieldBitfield("BUSFAULTENA", 17, 1),
            FieldBitfield("USGFAULTENA", 18, 1),
        ]),
        RegisterDef("CFSR", "Configurable Fault Status Register", 0xE000ED28, 4),
        RegisterDef("MMFSR", "MemManage Fault Status Register", 0xE000ED28, 1, [
            FieldBitfield("IACCVIOL", 0, 1),
            FieldBitfield("DACCVIOL", 1, 1),
            FieldBitfield("MUNSTKERR", 3, 1),
            FieldBitfield("MSTKERR", 4, 1),
            FieldBitfield("MLSPERR", 5, 1),
            FieldBitfield("MMARVALID", 7, 1)
        ]),
        RegisterDef("MMFAR", "MemManage Fault Address Register",
                    0xE000ED34, 4),
        RegisterDef("BFSR", "BusFault Status Register", 0xE000ED29, 1, [
            FieldBitfield("IBUSERR", 0, 1),
            FieldBitfield("PRECISERR", 1, 1),
            FieldBitfield("IMPRECISERR", 2, 1),
            FieldBitfield("UNSTKERR", 3, 1),
            FieldBitfield("STKERR", 4, 1),
            FieldBitfield("LSPERR", 5, 1),
            FieldBitfield("BFARVALID", 7, 1)
        ]),
        RegisterDef("BFAR", "BusFault Address Register", 0xE000ED38, 4),
        RegisterDef("UFSR", "UsageFault Status Register", 0xE000ED2A, 2, [
            FieldBitfield("UNDEFINSTR", 0, 1),
            FieldBitfield("INVSTATE", 1, 1),
            FieldBitfield("INVPC", 2, 1),
            FieldBitfield("NOCP", 3, 1),
            FieldBitfield("UNALIGNED", 8, 1),
            FieldBitfield("DIVBYZERO", 9, 1)
        ]),
        RegisterDef("HFSR", "HardFault Status Register", 0xE000ED2C, 4, [
            FieldBitfield("VECTTBL", 1, 1),
            FieldBitfield("FORCED", 30, 1),
            FieldBitfield("DEBUGEVT", 31, 1)
        ]),
        RegisterDef("AFSR", "Auxiliary Fault Status Register", 0xE000ED3C, 4)
    ]

    def __init__(self):
        super().__init__('arm scb', gdb.COMMAND_USER)
        self.parser = argparse.ArgumentParser(
            description="Inspect SCB - System Control Block"
        )
        self.parser.add_argument(
            '-d', '--descr',
            dest='descr',
            action='store_const',
            const=True,
            default=False,
            help="Include description"
        )

    def invoke(self, argument, from_tty):
        argv = gdb.string_to_argv(argument)
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            # We're running argparse in gdb, don't exit just return
            return

        inf = gdb.selected_inferior()
        for reg in self.regs:
            reg.dump(inf, args.descr)



# SPDX-FileCopyrightText: 2023 Max Sikström
# SPDX-License-Identifier: MIT

# Copyright © 2023 Max Sikström
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

def read_reg(inf, addr, len):
    bs = inf.read_memory(addr, len).tobytes()
    return sum(v << (8*i) for i, v in enumerate(bs))


def format_int(val, bits):
    chars = (bits + 3) // 4
    return (("0" * chars) + ("%x" % (val,)))[-chars:]


class RegisterDef:
    def __init__(self, name, descr, addr, size, fields=[]):
        self.name = name
        self.descr = descr
        self.addr = addr
        self.size = size
        self.fields = fields

    def dump(self, inf, include_descr=True):
        m_int = read_reg(inf, self.addr, self.size)

        if self.descr and include_descr:
            descr = (" "*18 + "// " + self.descr)
        else:
            descr = ""

        print("%-32s = %8s %s" %
              (self.name, format_int(m_int, self.size * 8), descr)
              )

        for field in self.fields:
            if field.should_print(m_int):
                field.print(m_int, include_descr)


class Field:
    def __init__(self, name, descr=None):
        self.name = name
        self.descr = descr

    def should_print(self, value):
        return False

    def get_bitmask(self, value):
        return 0

    def get_print_value(self, value):
        return None

    def print(self, value, include_descr=True):
        bitmask = self.get_bitmask(value)
        print_value = self.get_print_value(value)

        if self.descr and include_descr:
            descr = (" // " + self.descr)
        else:
            descr = ""

        print(
            "    %-28s   %8s - %-15s%s" %
            (
                self.name,
                format_int(bitmask, 32),
                print_value,
                descr
            )
        )


class FieldBitfield(Field):
    def __init__(self, name, bit_offset, bit_width, descr=None):
        super().__init__(name, descr)
        self.bit_offset = bit_offset
        self.bit_width = bit_width

    def get_bitmask(self, value):
        return value & (((1 << self.bit_width)-1) << self.bit_offset)

    def get_value(self, value):
        return (value >> self.bit_offset) & ((1 << self.bit_width)-1)

    def get_print_value(self, value):
        return format_int(self.get_value(value), self.bit_width)

    def should_print(self, value):
        return self.get_value(value) != 0


class FieldBitfieldEnum(FieldBitfield):
    def __init__(self, name, bit_offset, bit_width, enum_values, descr=None):
        super().__init__(name, bit_offset, bit_width, descr)
        self.enum_values = enum_values

    def get_enum_value(self, value):
        my_value = self.get_value(value)
        for enum_value in self.enum_values:
            if my_value == enum_value[0]:
                return enum_value
        return None

    def get_print_value(self, value):
        try:
            v, is_default, name, descr = self.get_enum_value(value)
            return name
        except:
            return format_int(self.get_value(value), self.bit_width)

    def should_print(self, value):
        try:
            v, is_default, name, descr = self.get_enum_value(value)
            return not is_default
        except:
            return True


class FieldBit(FieldBitfield):
    def __init__(self, name, bit, descr=None):
        super().__init__(name, bit, 1, descr)
        self.bit = bit



# SPDX-FileCopyrightText: 2023 Max Sikström
# SPDX-License-Identifier: MIT

# Copyright © 2023 Max Sikström
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import gdb
import argparse
from .common import *

from cmsis_svd.parser import SVDParser

devices = {}


class ArmToolsSVDList (gdb.Command):
    """List peripherals from a loaded SVD file"""

    def __init__(self):
        super().__init__('arm list', gdb.COMMAND_USER, gdb.COMPLETE_NONE, True)

        self.parser = argparse.ArgumentParser(
            description="Inspect a peripheral from a loaded SVD file"
        )
        self.parser.add_argument('device')

    def invoke(self, argument, from_tty):
        argv = gdb.string_to_argv(argument)
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            # We're running argparse in gdb, don't exit just return
            return

        device = devices[args.device]

        for peripheral in device.peripherals:
            print("%-10s @ 0x%08x" %
                  (peripheral.name, peripheral.base_address))


class ArmToolsSVDInspect (gdb.Command):
    """Dump peripherals specified by SVD file"""

    def __init__(self):
        super().__init__('arm inspect', gdb.COMMAND_USER, gdb.COMPLETE_NONE, True)

        self.parser = argparse.ArgumentParser(
            description="Inspect a peripheral from a loaded SVD file"
        )
        self.parser.add_argument('device')
        self.parser.add_argument('peripheral')
        self.parser.add_argument(
            '-d', '--descr',
            dest='descr',
            action='store_const',
            const=True,
            default=False,
            help="Include description"
        )

    def invoke(self, argument, from_tty):
        argv = gdb.string_to_argv(argument)
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            # We're running argparse in gdb, don't exit just return
            return

        device = devices[args.device]
        peripheral = None
        for p in device.peripherals:
            if p.name == args.peripheral:
                peripheral = p
        if peripheral is None:
            print("Unknown peripheral", args.peripheral)
            return

        inf = gdb.selected_inferior()

        for register in peripheral.registers:
            fields = []
            for field in register._fields:
                if field.is_enumerated_type:
                    fields.append(FieldBitfieldEnum(
                        field.name,
                        field.bit_offset,
                        field.bit_width,
                        [
                            (ev.value, ev.is_default, ev.name, ev.description)
                            for ev in field.enumerated_values
                        ],
                        field.description
                    ))
                else:
                    fields.append(FieldBitfield(
                        field.name,
                        field.bit_offset,
                        field.bit_width,
                        field.description
                    ))
            reg = RegisterDef(
                peripheral.name + "." + register.name,
                register.description,
                peripheral.base_address + register.address_offset,
                4,
                fields
            )
            reg.dump(inf, args.descr)


class ArmToolsSVDLoadFile (gdb.Command):
    """Load an SVD file from registry"""

    def __init__(self):
        super().__init__('arm loadfile', gdb.COMMAND_USER)

        self.parser = argparse.ArgumentParser(
            description="Load an SVD file given a path"
        )
        self.parser.add_argument('device')
        self.parser.add_argument('filename')

    def invoke(self, argument, from_tty):
        argv = gdb.string_to_argv(argument)
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            # We're running argparse in gdb, don't exit just return
            return

        parser = SVDParser.for_xml_file(args.filename)
        devices[args.device] = parser.get_device()


class ArmToolsSVDLoadDB (gdb.Command):
    """Load an SVD file from registry"""

    def __init__(self):
        super().__init__('arm loaddb', gdb.COMMAND_USER)

        self.parser = argparse.ArgumentParser(
            description="Load an SVD file from database. See cmsis-svd python package for a list"
        )
        self.parser.add_argument('device')
        self.parser.add_argument('vendor')
        self.parser.add_argument('filename')

    def invoke(self, argument, from_tty):
        argv = gdb.string_to_argv(argument)
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            # We're running argparse in gdb, don't exit just return
            return

        parser = SVDParser.for_packaged_svd(args.vendor, args.filename)
        devices[args.device] = parser.get_device()
